use crate::prelude::*;

use anyhow::Result;

use std::marker::PhantomData;
use std::slice;

#[derive(Debug)]
pub struct DescriptorWrite<'handle> {
    lifetime: &'handle PhantomData<()>,
    binding: u32,
    descriptor_type: VkDescriptorType,
    inner: InnerWrite,
}

#[derive(Debug)]
enum InnerWrite {
    Buffers(Vec<VkDescriptorBufferInfo>),
    Images(Vec<VkDescriptorImageInfo>),
    AS(
        (
            VkWriteDescriptorSetAccelerationStructureKHR,
            Vec<VkAccelerationStructureKHR>,
        ),
    ),
}

impl<'device, 'handle> DescriptorWrite<'handle> {
    pub fn uniform_buffers<T: ReprC + Send + Sync + 'static>(
        binding: u32,
        buffers: &[&'handle Buffer<'device, T>],
    ) -> Self {
        DescriptorWrite {
            binding,
            descriptor_type: VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
            inner: InnerWrite::Buffers(
                buffers
                    .iter()
                    .map(|buffer| VkDescriptorBufferInfo {
                        buffer: buffer.vk_handle(),
                        offset: 0,
                        range: buffer.byte_size(),
                    })
                    .collect(),
            ),
            lifetime: &PhantomData,
        }
    }

    pub fn storage_buffers<T: ReprC + Send + Sync + 'static>(
        binding: u32,
        buffers: &[&'handle Buffer<'device, T>],
    ) -> Self {
        DescriptorWrite {
            binding,
            descriptor_type: VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
            inner: InnerWrite::Buffers(
                buffers
                    .iter()
                    .map(|buffer| VkDescriptorBufferInfo {
                        buffer: buffer.vk_handle(),
                        offset: 0,
                        range: buffer.byte_size(),
                    })
                    .collect(),
            ),
            lifetime: &PhantomData,
        }
    }

    pub fn combined_samplers(binding: u32, images: &[&'handle Image<'device>]) -> Self {
        DescriptorWrite {
            binding,
            descriptor_type: VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
            inner: InnerWrite::Images(
                images
                    .iter()
                    .map(|image| VkDescriptorImageInfo {
                        sampler: image
                            .sampler()
                            .as_ref()
                            .expect("image has no sampler attached")
                            .vk_handle(),
                        imageView: image.vk_handle(),
                        imageLayout: image.image_layout(),
                    })
                    .collect(),
            ),
            lifetime: &PhantomData,
        }
    }

    pub fn storage_images(binding: u32, images: &[&'handle Image<'device>]) -> Self {
        DescriptorWrite {
            binding,
            descriptor_type: VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
            inner: InnerWrite::Images(
                images
                    .iter()
                    .map(|image| VkDescriptorImageInfo {
                        sampler: VkSampler::NULL_HANDLE,
                        imageView: image.vk_handle(),
                        imageLayout: image.image_layout(),
                    })
                    .collect(),
            ),
            lifetime: &PhantomData,
        }
    }

    pub fn input_attachments(binding: u32, images: &[&'handle Image<'device>]) -> Self {
        DescriptorWrite {
            binding,
            descriptor_type: VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT,
            inner: InnerWrite::Images(
                images
                    .iter()
                    .map(|image| VkDescriptorImageInfo {
                        sampler: VkSampler::NULL_HANDLE,
                        imageView: image.vk_handle(),
                        imageLayout: image.image_layout(),
                    })
                    .collect(),
            ),
            lifetime: &PhantomData,
        }
    }

    pub fn acceleration_structures(
        binding: u32,
        acceleration_structures: &[&'handle AccelerationStructure<'device>],
    ) -> Self {
        let vk_as: Vec<VkAccelerationStructureKHR> = acceleration_structures
            .iter()
            .map(|a| a.vk_handle())
            .collect();

        let mut write = DescriptorWrite {
            binding,
            descriptor_type: VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR,
            inner: InnerWrite::AS((
                VkWriteDescriptorSetAccelerationStructureKHR::default(),
                vk_as,
            )),
            lifetime: &PhantomData,
        };

        if let InnerWrite::AS((vk_write_as, vk_as)) = &mut write.inner {
            vk_write_as.set_acceleration_structures(vk_as);
        }

        write
    }

    pub fn change_image_layout(mut self, image_layout: VkImageLayout) -> Self {
        if let InnerWrite::Images(ref mut infos) = self.inner {
            for info in infos {
                info.imageLayout = image_layout;
            }
        }

        self
    }

    fn vk_write(&self, write: &mut VkWriteDescriptorSet) {
        match &self.inner {
            InnerWrite::Buffers(buffer_infos) => {
                write.set_buffer_infos(buffer_infos);
            }
            InnerWrite::Images(image_infos) => {
                write.set_image_infos(image_infos);
            }
            InnerWrite::AS((as_write, _)) => {
                write.descriptorCount = as_write.accelerationStructureCount;
                write.chain(as_write);
            }
        }
    }
}

pub struct DescriptorSetBuilder<'a> {
    device: &'a Device,
    descriptor_pool: &'a DescriptorPool<'a>,
    variable_desc_counts: Vec<u32>,
    variable_descriptor_count: VkDescriptorSetVariableDescriptorCountAllocateInfoEXT,
}

impl<'a> DescriptorSetBuilder<'a> {
    pub fn set_variable_descriptor_counts(mut self, descriptor_counts: &[u32]) -> Self {
        self.variable_desc_counts = descriptor_counts.to_vec();

        self
    }

    pub fn allocate<'handle>(mut self) -> Result<DescriptorSet<'a, 'handle>> {
        let layout = self.descriptor_pool.vk_handle();

        let mut descriptor_set_ci = VkDescriptorSetAllocateInfo::new(
            self.descriptor_pool.vk_handle(),
            slice::from_ref(&layout),
        );

        if !self.variable_desc_counts.is_empty() {
            self.variable_descriptor_count
                .set_descriptor_counts(&self.variable_desc_counts);
            descriptor_set_ci.chain(&self.variable_descriptor_count);
        }

        let descriptor_set = self.device.allocate_descriptor_sets(&descriptor_set_ci)?[0];

        Ok(DescriptorSet {
            device: self.device,
            pool: self.descriptor_pool,
            descriptor_set,

            lifetimes: &PhantomData,
        })
    }
}

#[derive(Debug)]
pub struct DescriptorSet<'device, 'handle> {
    device: &'device Device,
    pool: &'device DescriptorPool<'device>,
    descriptor_set: VkDescriptorSet,

    lifetimes: &'handle PhantomData<()>,
}

impl<'device, 'handle> DescriptorSet<'device, 'handle> {
    pub(crate) fn builder(
        device: &'device Device,
        descriptor_pool: &'device DescriptorPool<'device>,
    ) -> DescriptorSetBuilder<'device> {
        DescriptorSetBuilder {
            device,
            descriptor_pool,
            variable_desc_counts: Vec::new(),
            variable_descriptor_count: VkDescriptorSetVariableDescriptorCountAllocateInfoEXT::new(
                &[],
            ),
        }
    }

    // TODO: add update function for VkCopyDescriptorSet
    pub fn update(&self, writes: &[DescriptorWrite<'handle>]) -> Result<()> {
        debug_assert!(!writes.is_empty());

        #[cfg(debug_assertions)]
        {
            assert!(writes.len() <= self.pool.descriptor_set_layout.bindings().len());

            for binding in self.pool.descriptor_set_layout.bindings().iter() {
                if let Some(write) = writes.iter().find(|write| write.binding == binding.binding) {
                    assert_eq!(write.descriptor_type, binding.desc_type);
                }
            }
        }

        let mut vk_writes = Vec::new();

        for write in writes {
            let mut write_desc = VkWriteDescriptorSet::new(
                self.descriptor_set,
                write.binding,
                0,
                write.descriptor_type,
            );

            write.vk_write(&mut write_desc);

            vk_writes.push(write_desc);
        }

        self.device
            .update_descriptor_sets(vk_writes.as_slice(), &[]);

        Ok(())
    }
}

impl<'device, 'handle> VulkanDevice for DescriptorSet<'device, 'handle> {
    fn device(&self) -> &Device {
        &self.device
    }
}

impl_vk_handle!(
    DescriptorSet<'device, 'handle>,
    VkDescriptorSet,
    descriptor_set
);

impl<'device, 'handle> VkHandle<VkDescriptorSetLayout> for DescriptorSet<'device, 'handle> {
    fn vk_handle(&self) -> VkDescriptorSetLayout {
        self.pool.vk_handle()
    }
}

impl<'a, 'device, 'handle> VkHandle<VkDescriptorSetLayout> for &'a DescriptorSet<'device, 'handle> {
    fn vk_handle(&self) -> VkDescriptorSetLayout {
        self.pool.vk_handle()
    }
}

impl<'device, 'handle> Drop for DescriptorSet<'device, 'handle> {
    fn drop(&mut self) {
        if let Err(error) = self
            .device
            .free_descriptor_sets(self.pool.vk_handle(), &[self.descriptor_set])
        {
            println!("{}", error);
        }
    }
}

#[cfg(test)]
mod test {
    use crate::prelude::*;
    use anyhow::Result;

    #[test]
    fn create_multiple_sets_from_one_pool() -> Result<()> {
        const DESCRIPTOR_COUNT: u32 = 2;

        let (device, _queue) = crate::create_vk_handles()?;

        let descriptor_layout = DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            .build(&device)?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(&descriptor_layout)
            .set_descriptor_set_count(DESCRIPTOR_COUNT)
            .build(&device)?;

        let descriptors: Vec<DescriptorSet<'_, '_>> = (0..DESCRIPTOR_COUNT)
            .map(|_| {
                let set = descriptor_pool.prepare_set().allocate();

                assert!(set.is_ok(), "{}", set.err().unwrap());

                set.unwrap()
            })
            .collect();

        assert_eq!(descriptors.len(), DESCRIPTOR_COUNT as usize);

        Ok(())
    }
}
